/***********************************************************
*                       pC/TFS 2.38b                       *
*                 a FREE - Tiny File System                *
*            by Frank Goetze  -  www.embedded-os.de        *
************************************************************
*                    TFS_Port_SAMD21.c                     *
*                     SAMD21 NOR Port                      *
************************************************************
* Revision-Info:                                           *
*                                                          *
***********************************************************/
#include "../../inc/SAMD21/TFS_Port_SAMD21.h"

/*
************************************************************
*                      HW-specifics
************************************************************
*/
#define TFSport_NOR_PAGESIZE    256

// this version works in the same bank of code, so the HW-specific erase and write must be run in RAM !!!

extern U08 TFSport_erase_page_RAMfunct(U08 OS_HUGE *page);
extern U08 TFSport_write_page_RAMfunct(U08 OS_HUGE *dest, OS_CONST U08 OS_HUGE *src);

static const U32 TFSport_SAMD_page[] = { 8, 16, 32, 64, 128, 256, 512, 1024};
volatile U32 TFSport_SAMD_pagesize = 0;
volatile U32 TFSport_SAMD_rowsize  = 0;

U08 TFSport_erase_page(U08 OS_HUGE *page_addr)
{
#if OS_CRITICAL_SR
    OS_STK_TYPE  OS_sr;
#endif
    OS_ENTER_CRITICAL();
    TFSport_erase_page_RAMfunct(page_addr);
    OS_EXIT_CRITICAL();
    return TFS_NO_ERR;
}

U08 TFSport_write_page(U08 OS_HUGE *page_addr, OS_CONST U08 OS_HUGE* src)
{
#if OS_CRITICAL_SR
    OS_STK_TYPE  OS_sr;
#endif
    U32 i = 0;

    while(i < TFSport_SAMD_rowsize) {
        OS_ENTER_CRITICAL();
        TFSport_write_page_RAMfunct(&page_addr[i], &src[i]);
        OS_EXIT_CRITICAL();
        i += TFSport_SAMD_pagesize; 
    }
    return TFS_NO_ERR;
}

/*
************************************************************
*                    INIT NOR-MEMORY
************************************************************
*/
__align(4) U08 TFSport_page[TFSport_NOR_PAGESIZE]; // must be 32bit aligned

U08 TFSPort_Init(void)
{
#if OS_CRITICAL_SR
    OS_STK_TYPE  OS_sr;
#endif

    OS_ENTER_CRITICAL();
    TFSport_SAMD_pagesize = TFSport_SAMD_page[NVMCTRL->PARAM.bit.PSZ];
    TFSport_SAMD_rowsize  = TFSport_SAMD_pagesize * 4;
    OS_EXIT_CRITICAL();
    if((TFSport_SAMD_pagesize * NVMCTRL->PARAM.bit.NVMP) < TFS_SIZE) return(TFS_MEM_ERR);
    if(TFSport_SAMD_rowsize > TFSport_NOR_PAGESIZE) return(TFS_MEM_ERR);
    return(TFS_NO_ERR);
}

/*
************************************************************
*                  WRITE INTO NOR-MEMORY
************************************************************
*/
U08 TFSPort_Write(U08 OS_HUGE *dest, OS_CONST U08 OS_HUGE *source, TFS_LONG length)
{
    U08     *fdest, *fsrc;
    U32     segoffs, i, im;
    U08     ret;

    if((source >= (U08 OS_HUGE *)TFS_START) && (source < ((U08 OS_HUGE *)(TFS_START) + TFS_SIZE)))
        return(TFS_WR_PTR);                                // ERROR - source is in TFS-area
    if((dest < (U08 OS_HUGE *)TFS_START) || ((dest + length) > ((U08 OS_HUGE *)(TFS_START) + TFS_SIZE)))
        return(TFS_WR_PTR);                                // ERROR - destination is not in TFS-area
    segoffs =                 (U32)dest &  (TFSport_SAMD_rowsize - 1);
    fdest   = (U08 OS_HUGE *)((U32)dest & ~(TFSport_SAMD_rowsize - 1));
    while(length) {
        if ((segoffs > 0) || (length < TFSport_SAMD_rowsize)) {
            for (i = 0; i < TFSport_SAMD_rowsize; i += sizeof(U32)) *(U32 *)(&TFSport_page[i]) = *(U32 *)(&fdest[i]);
            im = ((segoffs + length) > TFSport_SAMD_rowsize)? TFSport_SAMD_rowsize : segoffs + length;
            for (i = segoffs; i < im; i++) TFSport_page[i] = *source++;
            length -= (im - segoffs);
            fsrc    = TFSport_page;
        } else {
            if (((U32)source & 0x03) != 0) {
                for (i = 0; i < TFSport_SAMD_rowsize; i++) TFSport_page[i] = source[i];
                fsrc    = TFSport_page;
            } else {
                fsrc    = source;
            }
            length -= TFSport_SAMD_rowsize;
            source += TFSport_SAMD_rowsize;
        }
        segoffs = 0;
        ret = TFSport_erase_page(fdest);
        if (ret != TFS_NO_ERR) return ret;
        ret = TFSport_write_page(fdest, fsrc);
        if (ret != TFS_NO_ERR) return ret;
        fdest += TFSport_SAMD_rowsize;
    }
    return(TFS_NO_ERR);
}

/*
************************************************************
*                      FILL NOR-MEMORY
************************************************************
*/
U08 TFSPort_Fill(U08 OS_HUGE *dest, U08 value, TFS_LONG length)
{
    U08     *fdest;
    U32     segoffs, i, im;
    U08     ret;

    if((dest < (U08 OS_HUGE *)TFS_START) || ((dest + length) > ((U08 OS_HUGE *)(TFS_START) + TFS_SIZE)))
        return(TFS_WR_PTR);                                // ERROR - destination is not in TFS-area
    segoffs =                 (U32)dest &  (TFSport_SAMD_rowsize - 1);
    fdest   = (U08 OS_HUGE *)((U32)dest & ~(TFSport_SAMD_rowsize - 1));
    while(length) {
        if ((segoffs > 0) || (length < TFSport_SAMD_rowsize)) {
            for (i = 0; i < TFSport_SAMD_rowsize; i += sizeof(U32)) *(U32 *)(&TFSport_page[i]) = *(U32 *)(&fdest[i]);
            im = (TFSport_SAMD_rowsize < length)? TFSport_SAMD_rowsize : segoffs + length;
            for (i = segoffs; i < im; i++) TFSport_page[i] = value;
            length -= (im - segoffs);
        } else {
            for (i = 0; i < TFSport_SAMD_rowsize; i++) TFSport_page[i] = value;
            length -= TFSport_SAMD_rowsize;
        }
        segoffs = 0;
        ret = TFSport_erase_page(fdest);
        if (ret != TFS_NO_ERR) return ret;
        ret = TFSport_write_page(fdest, TFSport_page);
        if (ret != TFS_NO_ERR) return ret;
        fdest += TFSport_SAMD_rowsize;
    }
    return(TFS_NO_ERR);
}

